"""Helper functions that are not specifically KNX related"""

def tohex(byte_array):
    """Convert a byte array to a HEX string representation."""
    return "".join("%02x " % b for b in byte_array)


def ip_to_array(ipaddress):
    """Convert a string representing an IPv4 address to 4 bytes."""
    res = []
    for i in ipaddress.split("."):
        res.append(int(i))

    assert len(res) == 4
    return res


def int_to_array(i, length=2):
    """Convert an length byte integer to an array of bytes."""
    res = []
    for dummy in range(0, length):
        res.append(i & 0xff)
        i = i >> 8
    return reversed(res)



"""KNX core module

This implements some core KNX classes and methods.
"""
import re
from knxip.helper import tohex

E_NO_ERROR = 0x00
E_HOST_PROTOCOL_TYPE = 0x01
E_VERSION_NOT_SUPPORTED = 0x02
E_SEQUENCE_NUMBER = 0x04
E_CONNECTION_ID = 0x21
E_CONNECTION_TYPE = 0x22
E_CONNECTION_OPTION = 0x23
E_NO_MORE_CONNECTIONS = 0x24
E_DATA_CONNECTION = 0x26
E_KNX_CONNECTION = 0x27
E_TUNNELING_LAYER = 0x28


def parse_group_address(addr):
    """Parse KNX group addresses and return the address as an integer.

    This allows to convert x/x/x and x/x address syntax to a numeric
    KNX group address
    """

    if addr is None:
        raise KNXException("No address given")

    res = None

    if re.match('[0-9]+$', addr):
        res = int(addr)

    match = re.match("([0-9]+)/([0-9]+)$", addr)
    if match:
        main = match.group(1)
        sub = match.group(2)
        res = int(main) * 256 + int(sub)

    match = re.match("([0-9]+)/([0-9]+)/([0-9]+)$", addr)
    if match:
        main = match.group(1)
        middle = match.group(2)
        sub = match.group(3)
        res = int(main) * 256 * 8 + int(middle) * 256 + int(sub)

    if res is None:
        raise KNXException("Address {} does not match any address scheme".
                           format(addr))

    return res


class ValueCache(object):
    """A simple caching class based on dictionaries"""

    def __init__(self):
        """Initialize an empty cache"""
        self.values = {}

    def get(self, name):
        """Get a value for the given name from cache"""
        return self.values.get(name)

    def set(self, name, value):
        """Set the cached value for the given name"""
        old_val = self.values.get(name)
        if old_val != value:
            self.values[name] = value
            return True
        else:
            return False

    def clear(self):
        """Remove all cached entries."""
        self.values = {}


class KNXException(Exception):
    """Exception when handling KNX functionalities"""

    errorcode = 0

    def __init__(self, message, errorcode=0):
        """Initialize exception with the given error message and error code"""
        super(KNXException, self).__init__(message)
        self.errorcode = errorcode

    def __str__(self):
        """Return a human-readable representation of the exception"""
        msg = {
            E_NO_ERROR: "no error",
            E_HOST_PROTOCOL_TYPE: "protocol type error",
            E_VERSION_NOT_SUPPORTED: "version not supported",
            E_SEQUENCE_NUMBER: "invalid sequence number",
            E_CONNECTION_ID: "invalid connection id",
            E_CONNECTION_TYPE: "invalid connection type",
            E_CONNECTION_OPTION: "invalid connection option",
            E_NO_MORE_CONNECTIONS: "no more connection possible",
            E_DATA_CONNECTION: "data connection error",
            E_KNX_CONNECTION: "KNX connection error",
            E_TUNNELING_LAYER: "tunneling layer error",
        }

        return super().__str__() + " " + msg.get(self.errorcode,
                                                 "unknown error code")

#pylint: disable=too-many-instance-attributes
#pylint: disable=invalid-name
class KNXMessage(object):
    """This represents a message on the KNX bus."""

    repeat = 0
    priority = 3   # (0 = system, 1 - alarm, 2 - high, 3 - normal)
    src_addr = 0
    dst_addr = 0
    dt = 0
    routing = 1
    length = 1
    data = [0]
    multicast = 1

    def __init__(self):
        """Initialize an empty KNX message"""
        pass

    def sanitize(self):
        """Sanitize all fields of the KNX message."""
        self.repeat = self.repeat % 2
        self.priority = self.priority % 4
        self.src_addr = self.src_addr % 0x10000
        self.dst_addr = self.dst_addr % 0x10000
        self.multicast = self.multicast % 2
        self.routing = self.routing % 8
        self.length = self.length % 16
        for i in range(0, self.length - 1):
            self.data[i] = self.data[i] % 0x100

    def to_frame(self):
        """Convert the object to its frame format."""
        self.sanitize()
        res = []
        res.append((1 << 7) + (1 << 4) + (self.repeat << 5) +
                   (self.priority << 2))
        res.append(self.src_addr >> 8)
        res.append(self.src_addr % 0x100)
        res.append(self.dst_addr >> 8)
        res.append(self.dst_addr % 0x100)
        res.append((self.multicast << 7) + (self.routing << 4) + self.length)

        for i in range(0, self.length - 1):
            res.append(self.data[i])

        checksum = 0
        for i in range(0, 5 + self.length):
            checksum += res[i]

        res.append(checksum % 0x100)

        return bytearray(res)

    @classmethod
    def from_frame(cls, frame):
        """Create a KNXMessage object from the frame format."""
        message = cls()

        # Check checksum first
        checksum = 0
        for i in range(0, len(frame) - 1):
            checksum += frame[i]

        if (checksum % 0x100) != frame[len(frame) - 1]:
            raise KNXException('Checksum error in frame {}, '
                               'expected {} but got {}'
                               .format(tohex(frame), frame[len(frame) - 1],
                                       checksum % 0x100))

        message.repeat = (frame[0] >> 5) & 0x01
        message.priority = (frame[0] >> 2) & 0x03
        message.src_addr = (frame[1] << 8) + frame[2]
        message.dst_addr = (frame[3] << 8) + frame[4]
        message.multicast = (frame[5] >> 7)
        message.routing = (frame[5] >> 4) & 0x03
        message.length = frame[5] & 0x0f
        message.data = frame[6:-1]

        if len(message.data) + 1 != message.length:
            raise KNXException(
                'Frame {} has not the correct length'.format(tohex(frame)))

        return message



"""Module to connect to a KNX bus using a KNX/IP tunnelling interface.
"""
import socket
import threading
import logging
import time
import queue as queue
import socketserver as SocketServer

from knxip.core import KNXException, ValueCache, E_NO_ERROR
from knxip.helper import int_to_array, ip_to_array
from knxip.gatewayscanner import GatewayScanner


class KNXIPFrame():
    """Representation of a KNX/IP frame."""

    SEARCH_REQUEST = 0x0201
    SEARCH_RESPONSE = 0x0202
    DESCRIPTION_REQUEST = 0x0203
    DESCRIPTION_RESPONSE = 0x0204
    CONNECT_REQUEST = 0x0205
    CONNECT_RESPONSE = 0x0206
    CONNECTIONSTATE_REQUEST = 0x0207
    CONNECTIONSTATE_RESPONSE = 0x0208
    DISCONNECT_REQUEST = 0x0209
    DISCONNECT_RESPONSE = 0x020a
    DEVICE_CONFIGURATION_REQUEST = 0x0310
    DEVICE_CONFIGURATION_ACK = 0x0111
    TUNNELING_REQUEST = 0x0420
    TUNNELLING_ACK = 0x0421
    ROUTING_INDICATION = 0x0530
    ROUTING_LOST_MESSAGE = 0x0531

    DEVICE_MGMT_CONNECTION = 0x03
    TUNNEL_CONNECTION = 0x04
    REMLOG_CONNECTION = 0x06
    REMCONF_CONNECTION = 0x07
    OBJSVR_CONNECTION = 0x08

    # CONNECTIONSTATE_RESPONSE Status Codes
    # 3.8.2 - 7.8.4
    E_DATA_CONNECTION = 0x26
    E_CONNECTION_ID = 0x21
    E_KNX_CONNECTION = 0x27

    # Generic Response Status Code
    E_NO_ERROR = 0x0

    body = None

    def __init__(self, service_type_id):
        """Initalize an empty frame with the given service type."""
        self.service_type_id = service_type_id

    def to_frame(self):
        """Return the frame as an array of bytes."""
        return bytearray(self.header() + self.body)

    @classmethod
    def from_frame(cls, frame):
        """Initilize the frame object based on a KNX/IP data frame."""
        # TODO: Check length
        ipframe = cls(frame[2] * 256 + frame[3])
        ipframe.body = frame[6:]
        return ipframe

    def total_length(self):
        """Return the length of the frame (in bytes)."""
        return 6 + len(self.body)

    def header(self):
        """Return the frame header (as an array of bytes)."""
        total_length = self.total_length()
        res = [0x06, 0x10, 0, 0, 0, 0]
        res[2] = (self.service_type_id >> 8) & 0xff
        res[3] = (self.service_type_id >> 0) & 0xff
        res[4] = (total_length >> 8) & 0xff
        res[5] = (total_length >> 0) & 0xff
        return res


# pylint: disable=too-few-public-methods
class KNXTunnelingRequest:
    """Representation of a KNX/IP tunnelling request."""

    def __init__(self):
        """Initialize object."""
        self.seq = 0
        self.cemi = None
        self.channel = 0

    @classmethod
    def from_body(cls, body):
        """Create a tunnelling request from a given body of a KNX/IP frame."""
        # TODO: Check length
        request = cls()
        request.channel = body[1]
        request.seq = body[2]
        request.cemi = body[4:]
        return request


# pylint: disable=too-many-instance-attributes
class CEMIMessage():
    """Representation of a CEMI message."""

    CMD_GROUP_READ = 1
    CMD_GROUP_WRITE = 2
    CMD_GROUP_RESPONSE = 3
    CMD_UNKNOWN = 0xff

    code = 0
    ctl1 = 0
    ctl2 = 0
    src_addr = None
    dst_addr = None
    cmd = None
    tpci_apci = 0
    mpdu_len = 0
    data = [0]

    def __init__(self):
        """Initialize object."""
        pass

    @classmethod
    def from_body(cls, cemi):
        """Create a new CEMIMessage initialized from the given CEMI data."""
        # TODO: check that length matches
        message = cls()
        message.code = cemi[0]
        offset = cemi[1]

        message.ctl1 = cemi[2 + offset]
        message.ctl2 = cemi[3 + offset]

        message.src_addr = cemi[4 + offset] * 256 + cemi[5 + offset]
        message.dst_addr = cemi[6 + offset] * 256 + cemi[7 + offset]

        message.mpdu_len = cemi[8 + offset]

        tpci_apci = cemi[9 + offset] * 256 + cemi[10 + offset]
        apci = tpci_apci & 0x3ff

        # for APCI codes see KNX Standard 03/03/07 Application layer
        # table Application Layer control field
        if apci & 0x080:
            # Group write
            message.cmd = CEMIMessage.CMD_GROUP_WRITE
        elif apci == 0:
            message.cmd = CEMIMessage.CMD_GROUP_READ
        elif apci & 0x40:
            message.cmd = CEMIMessage.CMD_GROUP_RESPONSE
        else:
            message.cmd = CEMIMessage.CMD_UNKNOWN

        apdu = cemi[10 + offset:]
        if len(apdu) != message.mpdu_len:
            raise KNXException(
                "APDU LEN should be {} but is {}".format(
                    message.mpdu_len, len(apdu)))

        if len(apdu) == 1:
            message.data = [apci & 0x2f]
        else:
            message.data = cemi[11 + offset:]

        return message

    def init_group(self, dst_addr=1):
        """Initilize the CEMI frame with the given destination address."""
        self.code = 0x11
        # frametype 1, repeat 1, system broadcast 1, priority 3, ack-req 0,
        # confirm-flag 0
        self.ctl1 = 0xbc
        self.ctl2 = 0xe0  # dst addr type 1, hop count 6, extended frame format
        self.src_addr = 0
        self.dst_addr = dst_addr

    def init_group_write(self, dst_addr=1, data=None):
        """Initialize the CEMI frame for a group write operation."""
        self.init_group(dst_addr)

        # unnumbered data packet, group write
        self.tpci_apci = 0x00 * 256 + 0x80

        if data is None:
            self.data = [0]
        else:
            self.data = data

    def init_group_read(self, dst_addr=1):
        """Initialize the CEMI frame for a group read operation."""
        self.init_group(dst_addr)
        self.tpci_apci = 0x00  # unnumbered data packet, group read
        self.data = [0]

    def to_body(self):
        """Convert the CEMI frame object to its byte representation."""
        body = [self.code, 0x00, self.ctl1, self.ctl2,
                (self.src_addr >> 8) & 0xff, (self.src_addr >> 0) & 0xff,
                (self.dst_addr >> 8) & 0xff, (self.dst_addr >> 0) & 0xff,
               ]
        if (len(self.data) == 1) and ((self.data[0] & 3) == self.data[0]):
            # less than 6 bit of data, pack into APCI byte
            body.extend([1, (self.tpci_apci >> 8) & 0xff,
                         ((self.tpci_apci >> 0) & 0xff) + self.data[0]])
        else:
            body.extend([1 + len(self.data), (self.tpci_apci >> 8) &
                         0xff, (self.tpci_apci >> 0) & 0xff])
            body.extend(self.data)

        return body

    def __str__(self):
        """Return a human readable string for debugging."""
        cmd = "??"
        if self.cmd == self.CMD_GROUP_READ:
            cmd = "RD"
        elif self.cmd == self.CMD_GROUP_WRITE:
            cmd = "WR"
        elif self.cmd == self.CMD_GROUP_RESPONSE:
            cmd = "RS"
        return "{0:x}->{1:x} {2} {3}".format(
            self.src_addr, self.dst_addr, cmd, self.data)


class KNXIPTunnel():
    """A connection to a KNX/IP tunnelling interface."""

    data_server = None
    control_socket = None
    channel = None
    seq = 0
    data_handler = None
    result_queue = None
    notify = None
    address_listeners = {}

    def __init__(self, ip, port=3671, valueCache=None):
        """Initialize the connection to the given host/port

        Initialized the connection, but does not connect.
        """
        self.remote_ip = ip
        self.remote_port = port
        self.discovery_port = None
        self.data_port = None
        self.connected = False
        self.result_queue = queue.Queue()
        self.ack_semaphore = threading.Semaphore(0)
        self.conn_state_ack_semaphore = threading.Semaphore(0)
        if valueCache is None:
            self.value_cache = ValueCache()
        else:
            self.value_cache = valueCache
        self.connection_state = 0
        self.keepalive_thread = threading.Thread(target=self.keepalive,
                                                 args=())
        self.keepalive_thread.daemon = True
        self.keepalive_thread.start()

    def __del__(self):
        """Make sure an open tunnel connection will be closed"""
        self.disconnect()

    def keepalive(self):
        """Background method that makes sure the connection is still open."""
        while True:
            if self.connected:
                self.check_connection_state()
            time.sleep(60)

    def connect(self, timeout=2):
        """Connect to the KNX/IP tunnelling interface.

        If the remote address is "0.0.0.0", it will use the Gateway scanner
        to automatically detect a KNX gateway and it will connect to it if one
        has been found.

        Returns true if a connection could be established, false otherwise
        """

        if self.connected:
            logging.info("KNXIPTunnel connect request ignored, "
                         "already connected")
            return True

        if self.remote_ip == "0.0.0.0":
            scanner = GatewayScanner()
            try:
                ipaddr, port = scanner.start_search()
                logging.info("Found KNX gateway %s/%s", ipaddr, port)
                self.remote_ip = ipaddr
                self.remote_port = port
            except TypeError:
                logging.error("No KNX/IP gateway given and no gateway "
                              "found by scanner, aborting %s")

        # Clean up cache
        self.value_cache.clear()

        # Find my own IP
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect((self.remote_ip, self.remote_port))
        local_ip = sock.getsockname()[0]

        if self.data_server:
            logging.info("Data server already running, not starting again")
        else:
            self.data_server = DataServer((local_ip, 0),
                                          DataRequestHandler,
                                          self)
            dummy_ip, self.data_port = self.data_server.server_address
            data_server_thread = threading.Thread(
                target=self.data_server.serve_forever)
            data_server_thread.daemon = True
            data_server_thread.start()
            logging.debug(
                "Started data server on UDP port %s", self.data_port)

        self.control_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.control_socket.bind((local_ip, 0))
        self.control_socket.settimeout(timeout)

        # Connect packet
        frame = KNXIPFrame(KNXIPFrame.CONNECT_REQUEST)

        # Control endpoint
        body = []
        body.extend([0x08, 0x01])  # length 8 bytes, UPD
        dummy_ip, port = self.control_socket.getsockname()
        body.extend(ip_to_array(local_ip))
        body.extend(int_to_array(port, 2))

        # Data endpoint
        body.extend([0x08, 0x01])  # length 8 bytes, UPD
        body.extend(ip_to_array(local_ip))
        body.extend(int_to_array(self.data_port, 2))

        #
        body.extend([0x04, 0x04, 0x02, 0x00])
        frame.body = body

        try:
            self.control_socket.sendto(bytes(frame.to_frame()),
                                       (self.remote_ip, self.remote_port))
            received = self.control_socket.recv(1024)
        except socket.error:
            self.control_socket = None
            logging.error("KNX/IP gateway did not respond to connect request")
            return False

        # Check if the response is an TUNNELING ACK
        r_sid = received[2] * 256 + received[3]

        if r_sid == KNXIPFrame.CONNECT_RESPONSE:
            self.channel = received[6]
            status = received[7]
            if status == 0:
                hpai = received[8:10]
                logging.debug("Connected KNX IP tunnel " +
                              "(Channel: {}, HPAI: {} {})".format(
                                  self.channel, hpai[0], hpai[1]))
            else:
                logging.error("KNX IP tunnel connect error:" +
                              "(Channel: {}, Status: {})".format(
                                  self.channel, status))
                return False

        else:
            logging.error(
                "Could not initiate tunnel connection, STI = {0:%s}", r_sid)
            return False

        self.connected = True

        return True

    def disconnect(self):
        """Disconnect an open tunnel connection"""
        if self.channel:
            logging.debug("Disconnecting KNX/IP tunnel...")

            frame = KNXIPFrame(KNXIPFrame.DISCONNECT_REQUEST)
            frame.body = self.hpai_body()

            # TODO: Glaube Sequence erhöhen ist nicht notwendig im Control
            # Tunnel beim Disconnect???
            if self.seq < 0xff:
                self.seq += 1
            else:
                self.seq = 0

            self.control_socket.sendto(
                bytes(frame.to_frame()), (self.remote_ip, self.remote_port))
            # TODO: Impelement the Disconnect_Response Handling from Gateway
            # Control Channel > Client Control Channel

        else:
            logging.debug("Disconnect - no connection, nothing to do")

        self.connected = False

    def check_connection_state(self):
        """Check the state of the connection using connection state request.

        This sends a CONNECTION_STATE_REQUEST. This method will only return
        True, if the connection is established and no error code is returned
        from the KNX/IP gateway
        """
        if not self.connected:
            self.connection_state = -1
            return False

        frame = KNXIPFrame(KNXIPFrame.CONNECTIONSTATE_REQUEST)
        frame.body = self.hpai_body()

        # Send maximum 3 connection state requests with a 10 second timeout
        res = False
        self.connection_state = 0

        maximum_retry = 3
        for retry_counter in range(0, maximum_retry):
            logging.debug("Heartbeat: Send connection state request")

            # Suggestion:
            # Carve the Control Socket out of the KNXIPTunnel
            # Class and Public only the Send and Receive
            # function and Implement in there the Heartbeat so we
            # can block when other Functions want to send
            self.control_socket.settimeout(10)  # Kind of a quirks
            self.control_socket.sendto(bytes(frame.to_frame()),
                                       (self.remote_ip, self.remote_port))

            try:
                self.control_socket.sendto(bytes(frame.to_frame()),
                                           (self.remote_ip, self.remote_port))
                receive = self.control_socket.recv(1024)

            except socket.timeout:
                logging.info("Heartbeat: No response, Retry Counter %d/%d",
                             retry_counter, maximum_retry)
                break

            frame = KNXIPFrame.from_frame(receive)
            if frame.service_type_id == KNXIPFrame.CONNECTIONSTATE_RESPONSE:
                if frame.body[1] == KNXIPFrame.E_NO_ERROR:
                    logging.debug("Heartbeat: Successful")
                    res = True
                    break
                if frame.body[1] == KNXIPFrame.E_CONNECTION_ID:
                    logging.error(
                        "Heartbeat: Response No active "\
                        "connection found for Channel:%d ", self.channel
                    )
                if frame.body[1] == KNXIPFrame.E_DATA_CONNECTION:
                    logging.error(
                        "Heartbeat: Response Data Connection Error Response "\
                        "for  Channel:%d ", self.channel
                    )
                if frame.body[1] == KNXIPFrame.E_DATA_CONNECTION:
                    logging.error(
                        "Heartbeat: Response KNX Sub Network Error Response "\
                        "for  Channel:%d ", self.channel
                    )
            else:
                logging.error("Heartbeat: Invalid Response!")

        if self.connection_state != 0:
            logging.info("Heartbeat: Connection state was %s",
                         self.connection_state)
            res = False

        if not res:
            if self.connection_state == 0:
                self.connection_state = -1
            self.disconnect()
            return False

        return True

    def hpai_body(self):
        """ Create a body with HPAI information.

        This is used for disconnect and connection state requests.
        """
        body = []
        # ============ IP Body ==========
        body.extend([self.channel])  # Communication Channel Id
        body.extend([0x00])  # Reserverd
        # =========== Client HPAI ===========
        body.extend([0x08])  # HPAI Length
        body.extend([0x01])  # Host Protocol
        # Tunnel Client Socket IP
        body.extend(ip_to_array(self.control_socket.getsockname()[0]))
        # Tunnel Client Socket Port
        body.extend(int_to_array(self.control_socket.getsockname()[1]))

        return body

    def send_tunnelling_request(self, cemi, auto_connect=True):
        """Sends a tunneling request based on the given CEMI data.

        This method does not wait for an acknowledge or result frame.
        """
        if self.data_server is None:
            if auto_connect:
                self.connect()
            else:
                raise KNXException("KNX tunnel not connected")

        frame = KNXIPFrame(KNXIPFrame.TUNNELING_REQUEST)
        # Connection header see KNXnet/IP 4.4.6 TUNNELLING_REQUEST
        body = [0x04, self.channel, self.seq, 0x00]
        if self.seq < 0xff:
            self.seq += 1
        else:
            self.seq = 0
        body.extend(cemi.to_body())
        frame.body = body
        self.data_server.socket.sendto(
            frame.to_frame(), (self.remote_ip, self.remote_port))

        # See KNX specification 3.8.4 chapter 2.6 "Frame confirmation"
        # Send KNX packet 2 times if not acknowledged and close
        # the connection if no ack is received
        res = self.ack_semaphore.acquire(blocking=True, timeout=1)
        # Resend package if not acknowledged after 1 seconds
        if not res:
            self.data_server.socket.sendto(
                frame.to_frame(), (self.remote_ip, self.remote_port))

            res = self.ack_semaphore.acquire(blocking=True, timeout=1)

        # disconnect and reconnect of not acknowledged
        if not res:
            self.disconnect()
            self.connect()

        return res

    def group_read(self, addr, use_cache=True, timeout=1):
        """Send a group read to the KNX bus and return the result."""
        if use_cache:
            res = self.value_cache.get(addr)
            if res:
                logging.debug(
                    "Got value of group address %s from cache: %s", addr, res)
                return res

        cemi = CEMIMessage()
        cemi.init_group_read(addr)
        # There might be old messages in the result quue, remove them
        self.result_queue.queue.clear()
        self.send_tunnelling_request(cemi)
        # Wait for the result
        try:
            res = self.result_queue.get(block=True, timeout=timeout)
        except queue.Empty:
            return None

        self.result_queue.task_done()
        return res

    def group_write(self, addr, data):
        """Send a group write to the given address.

        The method does not check if the address exists and the write request
        is valid.
        """
        cemi = CEMIMessage()
        cemi.init_group_write(addr, data)
        self.send_tunnelling_request(cemi)

    def group_toggle(self, addr, use_cache=True):
        """Toggle the value of an 1-bit group address.

        If the object has a value != 0, it will be set to 0, otherwise to 1
        """
        data = self.group_read(addr, use_cache)
        if len(data) != 1:
            problem = "Can't toggle a {}-octet group address {}".format(
                len(data), addr)
            logging.error(problem)
            raise KNXException(problem)

        if data[0] == 0:
            self.group_write(addr, [1])
        elif data[0] == 1:
            self.group_write(addr, [0])
        else:
            problem = "Can't toggle group address {} as value is {}".format(
                addr, data[0])
            logging.error(problem)
            raise KNXException(problem)

    def register_listener(self, address, func):
        """Adds a listener to messages received on a specific address

        If some KNX messages will be received from the KNX bus, this listener
        will be called func(address, data).
        There can be multiple listeners for a given address
        """
        try:
            listeners = self.address_listeners[address]
        except KeyError:
            listeners = []
            self.address_listeners[address] = listeners

        if not func in listeners:
            listeners.append(func)

        return True

    def unregister_listener(self, address, func):
        """Removes a listener function for a given address

        Remove the listener for the given address. Returns true if the listener
        was found and removed, false otherwise
        """
        listeners = self.address_listeners[address]
        if listeners is None:
            return False

        if func in listeners:
            listeners.remove(func)
            return True

        return False

    def received_message(self, address, data):
        """Process a message received from the KNX bus."""
        self.value_cache.set(address, data)
        if self.notify:
            self.notify(address, data)

        try:
            listeners = self.address_listeners[address]
        except KeyError:
            listeners = []

        for listener in listeners:
            listener(address, data)


class DataRequestHandler(SocketServer.BaseRequestHandler):
    """The class that handles incoming UDP packets from the KNX/IP tunnel."""

    def handle(self):
        """Process an incoming package."""
        data = self.request[0]
        sock = self.request[1]

        frame = KNXIPFrame.from_frame(data)

        if frame.service_type_id == KNXIPFrame.TUNNELING_REQUEST:
            req = KNXTunnelingRequest.from_body(frame.body)
            msg = CEMIMessage.from_body(req.cemi)
            send_ack = False

            tunnel = self.server.tunnel

            if msg.code == 0x29:
                # LData.req
                send_ack = True
            elif msg.code == 0x2e:
                # LData.con
                send_ack = True
            else:
                problem = "Unimplemented cEMI message code {}".format(msg.code)
                logging.error(problem)
                raise KNXException(problem)

            # Cache data
            if (msg.cmd == CEMIMessage.CMD_GROUP_WRITE) or (
                    msg.cmd == CEMIMessage.CMD_GROUP_RESPONSE):
                    # saw a value for a group address on the bus
                tunnel.received_message(msg.dst_addr, msg.data)

            # Put RESPONSES into the result queue
            if msg.cmd == CEMIMessage.CMD_GROUP_RESPONSE:
                tunnel.result_queue.put(msg.data)

            if send_ack:
                bodyack = [0x04, req.channel, req.seq, E_NO_ERROR]
                ack = KNXIPFrame(KNXIPFrame.TUNNELLING_ACK)
                ack.body = bodyack
                sock.sendto(ack.to_frame(), self.client_address)

        elif frame.service_type_id == KNXIPFrame.TUNNELLING_ACK:
            logging.debug("Received tunneling ACK")
            self.server.tunnel.ack_semaphore.release()
        elif frame.service_type_id == KNXIPFrame.DISCONNECT_RESPONSE:
            logging.debug("Disconnected")
            self.channel = None
            tunnel = self.server.tunnel
            tunnel.data_server.shutdown()
            tunnel.data_server = None
        elif frame.service_type_id == KNXIPFrame.CONNECTIONSTATE_RESPONSE:
            logging.debug("Connection state response")
            tunnel.connection_state = frame.body[2]
        else:
            logging.info(
                "Message type %s not yet implemented", frame.service_type_id)


class DataServer(SocketServer.ThreadingMixIn, SocketServer.UDPServer):
    """Server that handled the UDP connection to the KNX/IP tunnel."""

    def __init__(self, server_address, RequestHandlerClass, tunnel):
        super(DataServer, self).__init__(server_address, RequestHandlerClass)
        self.tunnel = tunnel
    


"""Conversions between KNX data types and native Python types"""

from datetime import time, date, datetime, timedelta
from knxip.core import KNXException


def float_to_knx2(floatval):
    """Convert a float to a 2 byte KNX float value"""

    if floatval < -671088.64 or floatval > 670760.96:
        raise KNXException("float {} out of valid range".format(floatval))

    floatval = floatval * 100

    i = 0
    for i in range(0, 15):
        exp = pow(2, i)
        if ((floatval / exp) >= -2048) and ((floatval / exp) < 2047):
            break

    if floatval < 0:
        sign = 1
        mantisse = int(2048 + (floatval / exp))
    else:
        sign = 0
        mantisse = int(floatval / exp)

    return [(sign << 7) + (i << 3) + (mantisse >> 8),
            mantisse & 0xff]


def knx2_to_float(knxdata):
    """Convert a KNX 2 byte float object to a float"""
    if len(knxdata) != 2:
        raise KNXException("Can only convert a 2 Byte object to float")

    data = knxdata[0] * 256 + knxdata[1]
    sign = data >> 15
    exponent = (data >> 11) & 0x0f
    mantisse = float(data & 0x7ff)
    if sign == 1:
        mantisse = -2048 + mantisse

    return mantisse * pow(2, exponent) / 100


def time_to_knx(timeval, dow=0):
    """Converts a time and day-of-week to a KNX time object"""

    knxdata = [0, 0, 0]
    knxdata[0] = ((dow & 0x07) << 5) + timeval.hour
    knxdata[1] = timeval.minute
    knxdata[2] = timeval.second
    return knxdata


def knx_to_time(knxdata):
    """Converts a KNX time to a tuple of a time object and the day of week"""

    if len(knxdata) != 3:
        raise KNXException("Can only convert a 3 Byte object to time")

    dow = knxdata[0] >> 5
    res = time(knxdata[0] & 0x1f, knxdata[1], knxdata[2])

    return [res, dow]


def date_to_knx(dateval):
    """Convert a date to a 3 byte KNX data array"""

    if (dateval.year < 1990) or (dateval.year > 2089):
        raise KNXException("Year has to be between 1990 and 2089")

    if dateval.year < 2000:
        year = dateval.year - 1900
    else:
        year = dateval.year - 2000

    return([dateval.day, dateval.month, year])


def knx_to_date(knxdata):
    """Convert a 3 byte KNX data object to a date"""

    if len(knxdata) != 3:
        raise KNXException("Can only convert a 3 Byte object to date")

    year = knxdata[2]
    if year >= 90:
        year += 1900
    else:
        year += 2000

    return date(year, knxdata[1], knxdata[0])


def datetime_to_knx(datetimeval, clock_synced_external=1):
    """Convert a Python timestamp to an 8 byte KNX time and date object"""

    res = [0, 0, 0, 0, 0, 0, 0, 0]
    year = datetimeval.year
    if (year < 1900) or (year > 2155):
        raise KNXException("Only years between 1900 and 2155 supported")
    res[0] = year - 1900
    res[1] = datetimeval.month
    res[2] = datetimeval.day
    res[3] = (datetimeval.isoweekday() << 5) + datetimeval.hour
    res[4] = datetimeval.minute
    res[5] = datetimeval.second
    if datetimeval.isoweekday() < 6:
        is_working_day = 1
    else:
        is_working_day = 0

    # DST starts last Sunday in March
    date1 = datetime(year, 4, 1)
    dston = date1 - timedelta(days=date1.weekday() + 1)
    # ends last Sunday in October
    date2 = datetime(year, 11, 1)
    dstoff = date2 - timedelta(days=date2.weekday() + 1)
    if dston <= datetimeval.replace(tzinfo=None) < dstoff:
        dst = 1
    else:
        dst = 0

    res[6] = (is_working_day << 6) + (1 << 5) + dst
    if clock_synced_external:
        res[7] = 128
    else:
        res[7] = 0

    return res


def knx_to_datetime(knxdata):
    """Convert a an 8 byte KNX time and date object to its components"""

    if len(knxdata) != 8:
        raise KNXException("Can only convert an 8 Byte object to datetime")

    year = knxdata[0] + 1900
    month = knxdata[1]
    day = knxdata[2]
    hour = knxdata[3] & 0x1f
    minute = knxdata[4]
    second = knxdata[5]

    return datetime(year, month, day, hour, minute, second)




"""

KNX IP Gateway Scanner

Search for an KNX IP Gateway in the network.

Reminder:
This Method is Blocking!

TODO:
    - Implement Return Object for the Device Information we get also from
    the Search Request Response

David Baumann(daBONDi@users.noreply.github.com)
"""

import asyncio
import logging
import struct

from knxip.helper import ip_to_array
from knxip.helper import int_to_array

logger = logging.getLogger(__name__)


class GatewayScanner:

    _default_timeout = 5  # Default Timeout for waiting for a Response
    _default_knx_port = 3671  # Default Port to Broadcast the searchrequest to
    # Default Multicast Address from KNX Assoc "224.0.23.12"
    _default_broadcast_address = "224.0.23.12"

    # Holder for the Transport Protocol Instance of the Broadcast listener
    _listener_transport = None
    # Holder for the Transport Protocol Instance of the Broadcast sender
    _broadcaster_transport = None

    _resolved_gateway_ip_address = None  # Gateway IP Address if found
    _resolved_gateway_ip_port = None  # Gateway Port if found

    _asyncio_loop = None

    def __init__(self, broadcast_source_ip_address="0.0.0.0",
                 broadcast_port=_default_knx_port,
                 broadcast_address=_default_broadcast_address,
                 timeout=_default_timeout):
        """
        Create a new Instance of a GatewayScanner Object

        :param broadcast_source_ip_address: IPv4 Address where we send the
            Search Request out and listen for response
        :param broadcast_port: UDP Port to send the Broadcast
            (Typical KNX IP Port: 3671)
        :param broadcast_address: Broadcast Address to send the
            KNX Searchrequest (Typical "224.0.23.12")
        :param timeout: Time in seconds to wait for an Response before
            returning None on the start_search method
        """
        self.__timeout = timeout
        self._broadcast_ip_address = broadcast_source_ip_address
        self._broadcast_address = broadcast_address
        self._broadcast_port = broadcast_port
        self._timeout = timeout
        pass

    def start_search(self):
        """
        Start the Gateway Search Request and return the address information

        :rtype: (string,int)
        :return: a tuple(string(IP),int(Port) when found or None when
            timeout occurs
        """

        self._asyncio_loop = asyncio.get_event_loop()

        """ Creating Broadcast Receiver """
        coroutine_listen = self._asyncio_loop.create_datagram_endpoint(
            lambda: self.KNXSearchBroadcastReceiverProtocol(
                self._process_response,
                self._timeout_handling,
                self._timeout,
                self._asyncio_loop
            ), local_addr=(self._broadcast_ip_address, 0)
        )
        self._listener_transport, listener_protocol = \
            self._asyncio_loop.run_until_complete(coroutine_listen)

        """ We are ready to fire the broadcast message """
        coroutine_broadcaster = self._asyncio_loop.create_datagram_endpoint(
            lambda: self.KNXSearchBroadcastProtocol(
                        self._asyncio_loop,
                        self._listener_transport.get_extra_info('sockname')
                        [1]),
            remote_addr=(self._broadcast_address, self._broadcast_port))

        self._broadcaster_transport, broadcast_protocol = \
            self._asyncio_loop.run_until_complete(coroutine_broadcaster)
        """ Waiting for all Broadcast receive or timeout """
        self._asyncio_loop.run_forever()

        """ We got Response or Timeout """
        if self._resolved_gateway_ip_address is None and \
           self._resolved_gateway_ip_port is None:
            logger.debug("Gateway not found!")
            return None
        else:
            logger.debug("Gateway found: {}:{}".format(
                self._resolved_gateway_ip_address,
                self._resolved_gateway_ip_port))

            return self._resolved_gateway_ip_address, \
                self._resolved_gateway_ip_port

    def _timeout_handling(self):
        """
        Timeout handler of the Broadcast Listener Socket
        """

        logger.error("Listener don't receive any packets, "
                     "timeout occur(Timeout in Seconds: {} !",
                     format(self._timeout))
        pass

    def _process_response(self, received_data):
        """
        Processing the incoming UDP Datagram from the Broadcast Socket

        :param received_data: UDP Datagram Package to Process
        :type received_data: Byte
        """
        r = bytearray(received_data)
        self._resolved_gateway_ip_address = str.format(
            "{}.{}.{}.{}", r[8], r[9], r[10], r[11])
        self._resolved_gateway_ip_port = struct.unpack(
            '!h', bytes(r[12:14]))[0]
        pass

    class KNXSearchBroadcastReceiverProtocol(asyncio.DatagramProtocol):

        def __init__(self,
                     broadcast_data_received_callback,
                     timeout_callback,
                     timeout_value,
                     loop):
            """
            Create a new KNX Search Broadcast Receiver Protocol Object

            :param broadcast_data_received_callback: function callback
                if data received
            :param timeout_callback: function callback if timeout occur
            :param timeout_value: Timeout in seconds when to stop waiting
            :param loop: Asyncio Loop to use
            """
            self.processing_data = broadcast_data_received_callback
            self.timeout_handler = timeout_callback
            self.timeout_in_seconds = timeout_value
            self.loop = loop
            self.transport = None

            self.h_timeout = asyncio.get_event_loop().call_later(
                self.timeout_in_seconds, self.timeout
            )

        def timeout(self):
            logger.error("Listener don't receive any packets, timeout!")
            self.transport.close()

        def connection_made(self, transport):
            logger.debug("Broadcast Receiver Started")
            self.transport = transport

        def datagram_received(self, data, addr):
            logger.debug(
                "Search Request Response received from {}:{}".
                format(addr[0], addr[1]))
            self.h_timeout.cancel()
            self.processing_data(data)
            self.transport.close()

        def error_received(self, exc):
            if exc:
                logger.error(
                    "Error on the broadcast receiver socket! {}".format(exc))

        def connection_lost(self, exc):
            if exc:
                logger.error(
                    "Error on the broadcast receiver socket! {}".format(exc))
            else:
                logger.debug("Closing broadcast receiver socket")

            self.loop.stop()
            super().connection_lost(exc)

    class KNXSearchBroadcastProtocol:

        def __init__(self, async_loop, listener_port):
            """
             Create a new KNX Search Broadcast Protocol Object

            :param async_loop: Asyncio Base Event Loop to use
            :type async_loop: BaseEventLoop
            """

            self.loop = async_loop
            self.transport = None
            self._listener_port = listener_port

        @staticmethod
        def _build_search_request_data(requestor_ipaddress, requestor_port):
            """
            Build a KNX IP Search Request Broadcast Package

            :param requestor_ipaddress: IP Address of the Search Request
                Response Receiver
            :param requestor_port: Port of the Search Request Response Receiver
            :return: Complete Search Request Broadcast Package
            :rtype: bytes
            """
            req = []
            req.extend([0x06])  # HeaderSize
            req.extend([0x10])  # KNXNetIP Version
            req.extend([0x02, 0x01])  # Search Reques°t
            req.extend([0x00, 0x0E])  # HEADER_SIZE_10 + sizeof(HPAI)
            # ==== Discovery Endpoint HPAI ====
            req.extend([0x08])  # Struct Length
            req.extend([0x01])  # Host Protocol Code = 0x01 = IPV4_UDP
            req.extend(ip_to_array(requestor_ipaddress))
            req.extend(int_to_array(requestor_port))
            return bytes(req)

        def connection_made(self, transport):
            self.transport = transport
            logger.debug("Broadcast Search Request Started")
            self.transport.sendto(
                self._build_search_request_data(
                    self.transport.get_extra_info(
                        'sockname')[0], self._listener_port
                )
            )
            logger.debug(
                "Search Request broadcast send, waiting for response ")

        def error_received(self, exc):
            logger.error("Error on Broadcast Socket")
